"""
Computation of distance matrix.
"""

import numpy as np
from scipy.spatial.distance import pdist, squareform
from sklearn.neighbors import RadiusNeighborsTransformer


def compute_dist_matrix(X, metric='inner_product', fill_diagonal=False):
    """
    Compute distance matrix between points and optionally fill diagonal elements
    with np.inf (may be convenient in some situation).

    Args:
        X (np.array): 2D array with data points.
        metric (str): Distance metric for scipy.spatial.distance.pdist or 'inner_product'.
            If 'inner_product' then use np.inner instead of pdist which is much faster.
            np.inner could be used instead of cosine distance for normalized vectors.
        fill_diagonal (bool): If True then fill diagonal with np.inf.
    """

    if X.ndim == 1:
        X = X[None, :]  # for correct work of distance computation

    if metric == 'inner_product':
        dist = 1 - np.inner(X, X)
    else:
        dist = pdist(X, metric=metric)
        # squareform converts emmpty dist array([]) to array([[0.]])
        # this behavior could break the code later
        dist = squareform(dist) if len(dist) > 0 else np.empty((0, 0))

    if fill_diagonal:
        np.fill_diagonal(dist, np.inf)

    return dist


def compute_sparse_dist_matrix(X, metric='cosine', max_distance=0.2):
    """
    Compute distance matrix in sparse csr format using sklearn RadiusNeighborsTransformer.
    Non-zero elements of matrix are elements for which distance is less than max_distance.

    Args:
        X (np.array): 2D array with data points.
        metric (str): Distance metric.
        max_distance (float): Maximum distance threshold.
    """

    transformer = RadiusNeighborsTransformer(mode='distance', algorithm='brute',
                                             metric=metric, radius=max_distance)

    return transformer.fit_transform(X)



import setuptools

with open("README.md", "r", encoding="utf-8") as readme_file:
    long_description = readme_file.read()

setuptools.setup(
    name="diameter-clustering",
    version="0.0.1",
    author="Anton Klenitskiy",
    author_email="ant-klen@yandex.ru",
    description="Clustering with maximum distance between points inside clusters",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/antklen/diameter-clustering",
    packages=['diameter_clustering'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    install_requires=[
        'numpy',
        'numpy_groupies',
        'scikit_learn',
        'scipy',
        'tqdm'
    ],
)



"""
Base classes for all clustering algorithms.
"""

import numpy as np

from .dist_matrix import compute_dist_matrix, compute_sparse_dist_matrix


class FitPredictMixin:
    """Mixin with fit_predict method."""

    def fit_predict(self, X):
        """Fit clustering from features or distance matrix and return cluster labels.

        Args:
            X (np.array or sparse matrix): Array with features or precomputed distance matrix,
                which could be in sparse matrix format.
        """

        self.fit(X)

        return self.labels_


class DistanceMatrixMixin:
    """Mixin with methods for working with distance matrix."""

    def _prepare_distance_matrix(self, X):
        """Prepare distance matrix.

        If self.precomputed_dist is True then do nothing, only check for square shape of X.
        Otherwise compute distance matrix regarding X as array of features. If self.sparse_dist
        is True then compute matrix in sparse format."""

        if not self.precomputed_dist:
            if self.sparse_dist:
                return compute_sparse_dist_matrix(X, metric=self.metric,
                                                  max_distance=self.max_distance)
            else:
                return compute_dist_matrix(X, metric=self.metric)

        if X.shape[0] != X.shape[1]:
            raise ValueError(f'Distance matrix should be square. Got matrix of shape {X.shape}.')

        return X

    def _slice_distance_matrix(self, dist, idx, indexes):
        """Get one row of distance matrix.
        Get distance between given point and several other points.

        Args:
            dist (np.array or csr_matrix): Distance matrix.
            idx (int): Index of given point.
            indexes (np.array): Indexes of other points.
        """

        if self.sparse_dist:
            current_dist = dist[idx, indexes].toarray()[0, :]
            current_dist[current_dist == 0] = np.inf
        else:
            current_dist = dist[idx, indexes]

        return current_dist
